/*************************************************************************
* This code was adapted from mdata.c written by Herbert Ward.  The changes
* reflect differences between the TPC and FTP architectures.  The structure
* follows a first (beta) draft of the FTP Raw Data Format (written by Jenn 
* Klay), which is very similar to the TPC Raw Data Format described in the
* document
* DAQ Raw Data Format, rev. 2.1
* ed. M. Levine   
* 3 June 1999
*
*  Adaptor:  Jenn Klay
*  Date:     10-Jun-1999
*
**************************************************************************/

/*******************************  includes  *****************/
#ifndef MDATA_H
#include "mdata.h"
#endif
#ifndef FTP_H
#include "FTP.h"
#endif

/*********************************************  functions  *************/
void InitGlobalsForFTPCPPR(void) {
  int ii; unsigned char *jelly;
  for(ii=0;ii<NUMDATAWORDSFTPCPPR;ii++) gFTPCPP[ii]=0xfc00fc00;
  jelly=(unsigned char*)&(gFTPCPP[0]);
  jelly[0]=(0.2)*PULSEHEIGHT;
  jelly[1]=(0.8)*PULSEHEIGHT;
  jelly[2]=1; /* BBB ???  See page 14, data word 1 of TPCCPPR. */
  jelly[3]=1; /* BBB ???  See page 14, data word 1 of TPCCPPR. */
}

void PrepareGlobalsForFTPCPPR(ulong whichPad,ulong nCenterbins,ulong *cbin) {
  ulong basicOffset,ii; ushort *peanutButter;
  int beg,end;
  if(whichPad>=PADSPERFTPSECTOR) ERR;
  basicOffset=32*whichPad+1; /* First word is diff for this bank. */
  for(ii=0;ii<nCenterbins;ii++) {
    if(ii+basicOffset>=NUMDATAWORDSFTPCPPR) ERR;
    if(ii+basicOffset< 0               ) ERR;
    peanutButter=(ushort*)(&(gFTPCPP[ii+basicOffset]));
    beg=cbin[ii]-PULSEWIDTH/2; if(beg<1) beg=1; if(beg>512) beg=512;
    end=cbin[ii]+PULSEWIDTH/2; if(end<1) end=1; if(end>512) end=512;
    peanutButter[0]=beg; peanutButter[1]=end;
  }
}

void WriteBodyFTPCPPR(char swapbytes,ulong junk) {
  write(gFd,gFTPCPP,4*NUMDATAWORDSFTPCPPR);
}

ulong FTPOoo(ulong xx) {
  ulong rv=1;
  if(xx>0) rv*=513; /* 512 time bins */
  if(xx>1) rv*=161; /* 160 pads in row */
  if(xx>2) rv*=3;   /* 2 padrows in sector */
  if(xx>3) rv*=61;  /* 60 sectors in detector (30 per chamber) */
  if(xx>4) rv*=3;   /* FTP = 2 */
  return rv;
}

void WriteBodyFTPP(char swapbytes,ulong eventNum0) { /* p9 */
  ulong uu,ii;
  if(swapbytes) ERR;
 
  /* FTPP word 1-4 = Offset/Length to chamber contributions */
  uu=0; for(ii=0;ii<4;ii++) write(gFd,&uu,4); /* Offsets and lengths
  for non-empty chambers are written later. (comment 77y)*/
  
}  

void WriteBodyFTPCHAP(char swapbytes,ulong eventNum1) { /* p9 */
  ulong uu,ii;
  if(swapbytes) ERR;
  
  /* FTPCHAP word 1-20 = Offset/Length to rcvboard contributions */
  uu=0; for(ii=0;ii<20;ii++) write(gFd,&uu,4); /* Offsets and lengths
  for non-empty rcvboards are written later. (comment 77z)*/
 
}      

void WriteBodyFTPRBP(char swapbytes,ulong notUsed) {
  ulong ii,uu=0;
  if(swapbytes) ERR;
  
  /* FTPRBP word 1-6 = Offset/Length to sector contributions */
  /* FTPRBP word 7-22 = Fiber Header information */
  for(ii=1;ii<=22;ii++) write(gFd,&uu,4);
}

void WriteBodyFTPAZIP(char swapbytes,ulong notUsed) { /* p9 */
  ulong uu,ii;
  if(swapbytes) ERR;
  
  /* FTPAZIP word 1-6 = Offset/Lengths to mezzanine board contributions */
  uu=0; for(ii=0;ii<6;ii++) write(gFd,&uu,4); /* Offsets and lengths
  for non-empty mezzanines are written later. (comment 77z)*/
 
}      

void WriteBodyFTPMZP(char swapbytes,ulong notUsed) {
  /* The length is variable (p 10), but we do no harm by writing
  ** all 24 words, and may avoid a bug later. */
  ulong ii,uu=0;
  if(swapbytes) ERR;
  for(ii=1;ii<=24;ii++) write(gFd,&uu,4);
}

void FillLongsForThisFTPPad(ulong nCenterbins,ulong *cbin,ulong *data) {
  /* INTEGERS SO THAT NEG OUT OF BOUNDS VALUES CAN BE SKIPPED. */ int jj,kk;
  unsigned char *buf; ulong nn; int cnt=0; 
  if(PULSEWIDTH%2!=1) ERR; /* must be odd so that centerbin is not ambiguous */
  nn=PULSEWIDTH/2;
  buf=(unsigned char*)data;
  for(jj=0;jj<512;jj++) { buf[jj]=0; }
    for(jj=0;jj<nCenterbins;jj++) { /* crude rectangular shape for pulses */
      for(kk=cbin[jj]-nn;kk<=cbin[jj]+nn;kk++) {
        if(kk<0||kk>=512) continue; /* time bucket number out of data range */
        if (++cnt<=PULSEWIDTH/2) {
  	  buf[kk]+=PULSEHEIGHT*(cnt)/nn;
        } else {
  	  buf[kk]+=PULSEHEIGHT*(PULSEWIDTH+1-cnt)/nn;
//          printf("adc value set to %d on bin %d\n", buf[kk],kk);
	}
      }
    }
}

void FillLongsForFTPPADK(ulong *data) { /* see nested struct page 12 */
  unsigned char *buf;
  ulong ii,rowIndex,padIndex;
  buf=(unsigned char*)data;
  for(ii=0;ii<192*4;ii++) {buf[ii]=0xff;}
	//  printf("writing info for %d pads\n", gNFTPpad);
  for(ii=0;ii<gNFTPpad;ii++) {
    rowIndex=ii*2; padIndex=ii*2+1;
    if(padIndex>=192*4) ERR;
    buf[rowIndex]=gFTPP[ii].row;
    buf[padIndex]=gFTPP[ii].pad;
	//    printf("entry %d row %d pad %d\n", ii, gFTPP[ii].row, gFTPP[ii].pad);
  }
	//  for(ii=0;ii<192*4;ii++)
	//   printf("-> buf[%d]=%d\n", ii, buf[ii]);
}

void WriteBodyFTPPADK(char swapbytes,ulong eventNum0) { /* see page 12 */
  ulong ii,uu,data[192];   
  if(swapbytes) ERR;
     
  /* TPCPADK word 1 */
  uu=512; /* see WriteBodyFTPADCR */ write(gFd,&uu,4);

  /* TPCPADK word 2-6 */
  uu=0; for(ii=2;ii<=6;ii++) write(gFd,&uu,4);
  
  FillLongsForFTPPADK(data);
  write(gFd,&data,4*192);  
}

void WriteBodyFTPADCR(char swapbytes,ulong eventAndSector) { /* see page 12 */
  ulong pad[PADSPERFTPSECTOR],row[PADSPERFTPSECTOR],iC,ii;
  ulong data[128]; /* 4 ADC values per word (ie, per long) = 512 total, p 12 */
  ulong cbin[CENTERBIN],nCenterbins,iL,ihit,event,sector,nlist=0;
  char already;
  event=eventAndSector/1000;
  sector=eventAndSector-1000*event;
  if(swapbytes) ERR;
  if(FTPADCR_FORMAT!=0) ERR;  /* compression not implemented, see page 12 */
  
  /* mk lst of padrows/pads 4 this sector (to support 2 hits on same pad) */
  /*for(ihit=0;ihit<gFTPEvent[event].nhit;ihit++) { 
     if(gFTPEvent[event].sector[ihit]!=sector) continue; 
     already=0; 
     for(iL=0;iL<nlist;iL++) { 
       if(gFTPEvent[event].row[ihit]==row[iL]&&gFTPEvent[event].pad[ihit]==pad[iL]) 
       { 
         already=7; break;   
       } 
     } 
     if(!already) { 
       if(nlist>=PADSPERFTPSECTOR) ERR; 
       row[nlist]=gFTPEvent[event].row[ihit]; 
       pad[nlist]=gFTPEvent[event].pad[ihit]; 
       nlist++; 
     } 
   } */

   /* loop over all pads because not written pads confuse 
      FTPV1P0_PADK_SR::initialize() This problem should really be corrected 
      in the readers... */ 
    nlist=PADSPERFTPSECTOR;
    for(iL=0;iL<nlist;iL++) {  
      row[iL]=(int) iL/(PADSPERFTPSECTOR/2)+1; 
      pad[iL]=iL - (row[iL]-1)*(PADSPERFTPSECTOR/2)+1; 
    } 
 
  /* now use the list (pad[] and row[]) */
  gNFTPpad=0;
  InitGlobalsForFTPCPPR();
  for(iL=0;iL<nlist;iL++) { /* one loop per pad */  
    /* Make list of centerbins from input file [see Usage()]. */
    nCenterbins=0;
    for(ihit=0;ihit<gFTPEvent[event].nhit;ihit++) {
      if(gFTPEvent[event].sector[ihit]!=sector) continue;
      if(gFTPEvent[event].row[ihit]!=row[iL]) continue;
      if(gFTPEvent[event].pad[ihit]!=pad[iL]) continue;
      if(nCenterbins>=CENTERBIN) ERR;
      cbin[nCenterbins++]=gFTPEvent[event].cbin[ihit];
    }
    if(gNFTPpad>=PADSPERFTPSECTOR) ERR; gFTPP[gNFTPpad].pad=pad[iL]; gFTPP[gNFTPpad].row=row[iL];
    gNFTPpad++;
    FillLongsForThisFTPPad(nCenterbins,cbin,data);
    PrepareGlobalsForFTPCPPR(iL,nCenterbins,cbin);
//    for(ii=0; ii<128; ii++)
// 	printf("pad %d data[%d]=%d\n",pad[iL], ii, data[ii]);
    write(gFd,data,512);
  }
  
  /* Fill out the rest of the bank with zeros. */
  for(ii=0;ii<128;ii++) data[ii]=0;
  for(ii=0;ii<320-nlist;ii++) write(gFd,data,512); 
}

/* comment 88n: These are non-zero only during the init, to reduce
** potential for confusion. */

void GetFTPHitsThisSector(ulong event,ulong sector1,
      ulong *row,ulong *pad,ulong *cbin,ulong *nhits) {
  ulong ii,cnt=0;
  for(ii=0;ii<gFTPEvent[event].nhit;ii++) {
    if(gFTPEvent[event].sector[ii]!=sector1) continue;
    if(cnt>=HITSPERSECTOR) ERR;
    row[cnt]=gFTPEvent[event].row[ii];
    pad[cnt]=gFTPEvent[event].pad[ii];
    cbin[cnt]=gFTPEvent[event].cbin[ii];
    cnt++;
  }
  *nhits=cnt;
}

void WriteBodyFTPSEQD(char swapbytes,ulong arg) {
  ulong which73x,cnt=0,event0,sector1,ii; size_t nwords;
  ulong row[HITSPERSECTOR],pad[HITSPERSECTOR],cbin[HITSPERSECTOR];
  ulong body[MEM],ihit,nhit,packcnt=0;
  ushort *packed; void *silenceCompiler;
  char lastHitThisPad,newMarker,x0=0,x1=0,x2=0;
  silenceCompiler=body; packed=silenceCompiler; /* data packed 2/word */
  for(ii=0;ii<MEM;ii++) body[ii]=0;
  event0=arg/1000; sector1=arg-1000*event0;
  GetFTPHitsThisSector(event0,sector1,row,pad,cbin,&nhit); gFTPOtherBank=0;
  for(ihit=0;ihit<nhit;ihit++) {
    if(row[ihit]<1||row[ihit]>2) ERR;
    if(row[ihit]==1&&!x0) {
      x0=7; gFTPADCX[0].oSEQD=Curr(I)-gPh+ihit;
    }
    if(row[ihit]==2&&!x1) {
      x1=7; gFTPADCX[1].oSEQD=Curr(I)-gPh+ihit;
    }
    newMarker=0;
    if(ihit==0||row[ihit]!=row[ihit-1]) { /* new padrow */ newMarker=7; }
    else if(pad[ihit]>pad[ihit-1]+1) { /* skipped a pad */ newMarker=7; }
    lastHitThisPad=0;
    if(ihit<nhit-1) {
      if(pad[ihit]!=pad[ihit+1]) lastHitThisPad=7;
      if(row[ihit]!=row[ihit+1]) lastHitThisPad=7;
    } else lastHitThisPad=7;
    PP"event=%d, sector=%02d, row=%02d, pad=%02d, cb=%03d, %d %d\n",
    event0,sector1,row[ihit],pad[ihit],cbin[ihit],newMarker,lastHitThisPad);
    if(newMarker) {
      if(packcnt>=2*MEM) ERR; packed[packcnt++]=Switch1(row[ihit],pad[ihit]);
    }
    if(packcnt>=2*MEM) ERR;
    packed[packcnt++]=Switch0(lastHitThisPad,cbin[ihit]); gFTPOtherBank++;
  }
  nwords=(packcnt+1)/2; if(nwords>=MEM) ERR; write(gFd,body,4*nwords);
  for(ii=0; ii<packcnt; ii++)
     printf("seq[%d]=%d\n", ii, (short)packed[ii]);

}

void WriteBodyFTPADCD(char swapbytes,ulong arg) { /* page 11 */
  ulong body[NMEM],row[HITSPERSECTOR],pad[HITSPERSECTOR],cbin[HITSPERSECTOR];
  ulong ii,nhit,ihit,event0,sector1,ndata=0; size_t nwords;
  unsigned char *data; void *b;
  char x0=0,x1=0,x2=0;
  event0=arg/1000; sector1=arg-1000*event0; b=body; data=b;
  GetFTPHitsThisSector(event0,sector1,row,pad,cbin,&nhit);
  for(ii=0;ii<NMEM;ii++) body[ii]=0;
  for(ihit=0;ihit<nhit;ihit++) {
    if(row[ihit]<1||row[ihit]>2) ERR;
    if(row[ihit]==1&&!x0) {
      x0=7; gFTPADCX[0].oADCD=Curr(I)-gPh+ihit; gFTPADCX[0].rowNumber=row[ihit];   
    }
    if(row[ihit]==2&&!x1) {
      x1=7; gFTPADCX[1].oADCD=Curr(I)-gPh+ihit; gFTPADCX[1].rowNumber=row[ihit];
    }
    for(ii=0;ii<PULSEWIDTH;ii++) {
      if(ndata>=NMEM*4) ERR;
       if(data[ndata]>=256-PULSEHEIGHT) ERR; 
       if(ii<=PULSEWIDTH/2) {
         data[ndata++]+=PULSEHEIGHT*(ii+1)/(PULSEWIDTH/2);
       }
       else {
         data[ndata++]+=PULSEHEIGHT*(PULSEWIDTH-ii)/(PULSEWIDTH/2);
       }
    }
  }
  if(ndata!=PULSEWIDTH*gFTPOtherBank) ERR;
  PP"bbb ndata=%d\n",ndata);
  nwords=(ndata+3)/4; if(nwords>=MEM) ERR; write(gFd,body,4*nwords);
}

void WriteBodyFTPADCX(char swapbytes,ulong arg) { /* page 13 */
  ulong val;
  int ii; /* int OK */
  for(ii=0;ii<3;ii++) {
    if(gFTPADCX[ii].oADCD==0) continue;
    val=gFTPADCX[ii].rowNumber; write(gFd,&val,4);
    val=gFTPADCX[ii].oADCD-40;  write(gFd,&val,4); /* 40 bytes in header */
    val=gFTPADCX[ii].oSEQD-40;  write(gFd,&val,4); /* 40 bytes in header */  
  }  
}

ulong NumFTPHits(ulong ev0,ulong sector1) {
  ulong rv=0; long ii;
  for(ii=gFTPEvent[ev0].nhit-1;ii>=0;ii--) {
    if(gFTPEvent[ev0].sector[ii]==sector1) rv++;
  }
  return rv;
}

void InitTheFTPComment73xGlobals(void) {
  int ii;
  for(ii=0;ii<2;ii++) { gFTPADCX[ii].oADCD=0; gFTPADCX[ii].oSEQD=0; }
}
