#include "mdata.h"

typedef struct {
  int L0Data[32];
  int L1Data[64];
  int L2Data[64];
  int RawCTB[64];
  int RawMWZ[36];
  int RawEMC[120];
} TRGDATA_T;

static TRGDATA_T gTrgd;
extern int gFd;

void InitTrigData(void)
{
  int lcv;
  for (lcv=0;lcv<32;lcv++)
    {
      gTrgd.L0Data[lcv]=0; gTrgd.L1Data[lcv]=0; gTrgd.L2Data[lcv]=0;
      gTrgd.RawMWZ[lcv]=0; gTrgd.RawCTB[lcv]=0; gTrgd.RawEMC[lcv]=0;
    }

  for (lcv=32;lcv<36;lcv++)
    {
      gTrgd.L1Data[lcv]=0; gTrgd.L2Data[lcv]=0;
      gTrgd.RawMWZ[lcv]=0; gTrgd.RawCTB[lcv]=0; gTrgd.RawEMC[lcv]=0;
    }

  for (lcv=36;lcv<64;lcv++)
    {
      gTrgd.L1Data[lcv]=0; gTrgd.L2Data[lcv]=0;
      gTrgd.RawCTB[lcv]=0; gTrgd.RawEMC[lcv]=0;
    }
  for (lcv=64;lcv<120;lcv++) gTrgd.RawEMC[lcv]=0;
}

int ReadTrigData(char *fname)
{
  FILE *tif;
  int ln=0,sect;
  char line[80],*cc;

  tif=fopen(fname,"r");
  if (!tif) {printf("ERR - Trigger Input file %s not found\n",fname);return 1;}
  while( fgets(line,80,tif) )
    {
      ln++;
      if(*line=='#') continue;
      cc=strtok(line," \n"); if (!cc) return 1;
      if (*cc=='L')
	{
	  switch (*(cc+1))
	    {
	    case '0':  
		{
		  sect=0;
		  while ( (cc=strtok(NULL," \n")) && (sect<32) )
		      gTrgd.L0Data[sect++]=atoi(cc);
		  break;
		}
	    case '1':
		{
		  sect=0;
		  while ( (cc=strtok(NULL," \n")) && (sect<64) )
		      gTrgd.L1Data[sect++]=atoi(cc);
		}
	    case '2':
		{
		  sect=0;
		  while ( (cc=strtok(NULL," \n")) && (sect<64) )
		      gTrgd.L2Data[sect++]=atoi(cc);
		  break;
		}
	    default:
	      {
		fclose(tif);
		return 1;
	      }
	    }
	}
      else
	{
	  switch (*cc)
	    {
	    case 'M':
	      {
                sect=0;
		if (*(cc+1)!='Z') return 1;
		while ( (cc=strtok(NULL," \n")) && (sect<36) )
		  gTrgd.RawMWZ[sect++]=atoi(cc);
		break;
	      }
	    case 'C':
	      {
                sect=0;
		if (*(cc+1)!='T') return 1;
		  while ( (cc=strtok(NULL," \n")) && (sect<64) )
		      gTrgd.RawCTB[sect++]=atoi(cc);
		break;
	      }
	    case 'E':
	      {
                sect=0;
		if (*(cc+1)!='C') return 1;
                   while ( (cc=strtok(NULL," \n")) && (sect<120) )
		      gTrgd.RawEMC[sect++]=atoi(cc);
		break;
	      }
	    default:
	      {
		printf("ERR - %s is not a trigger input file\n",fname);
		fclose(tif);
		return 1;
	      }
	    }
	}
    }
    fclose(tif);
}
void WriteBodyTRGD(char swapbytes,ulong arg)
{
  ulong uu;
  short ss1,ss2;
  char cc1,cc2;

  /* First write Event Descriptor */

  /* Word 1,2 MS and LS 32 bits of Bunch crossing number */
  uu=0x4868; write(gFd,&uu,4);
  uu=0x4869; write(gFd,&uu,4);
/* Word 3  2bitToken+2bitActionWord */
ss1=3;ss2=4;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 4 Output from last DSM+Address*/
  ss1=4;ss2=5;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 5 1b BusyBits + 1b flag + 2b TrgWord */
  cc1=5;cc2=6;ss1=(short)(cc2<<8)+(short)cc1;
  ss2=7;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 6 Npre+Npost */
  ss1=6;ss2=7;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);

  /* Write Trigger Summary */

  /* Word 7 Length of Trigger Summary/format */
  ss1=656; ss2=1; uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);

  /* Word 8 "L0"/Length of L0 data */
  cc1='L';cc2='0';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=128;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Words 9->40 L0 Data */
  for (ss1=0;ss1<32;ss1++) 
    {
      uu=gTrgd.L0Data[ss1]; write(gFd,&uu,4);
    }
  /* Word 41 "L1"/Length of L1 data */
  cc1='L';cc2='1';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=256;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Words 42->105 L1 Data */
  for (ss1=0;ss1<64;ss1++) 
    {
     uu=gTrgd.L1Data[ss1]; write(gFd,&uu,4);
    }
  /* Word 106 "L2"/Length of L2 data */
  cc1='L';cc2='2';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=256;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Words 107->170 L2 Data */
  for (ss1=0;ss1<64;ss1++) 
    {
       uu=gTrgd.L2Data[ss1]; write(gFd,&uu,4);
    }
  /* Write Raw Data */

  /* Word 171 Length of raw data block/format */
  ss1=896; ss2=1; uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);

  /* Word 172 "CT"/Length of CTB raw data */
  cc1='C';cc2='T';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=256;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 173->236 CTB RAW DATA */
  for (ss1=0;ss1<64;ss1++) 
    {
      uu=gTrgd.RawCTB[ss1]; write(gFd,&uu,4);
    }
  /* Word 237 "MZ"/Length of MWC/ZDC raw data */
  cc1='M';cc2='Z';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=144;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 238->273 MWC/ZDC RAW DATA */
  for (ss1=0;ss1<36;ss1++)
    {
      uu=gTrgd.RawMWZ[ss1]; write(gFd,&uu,4);
    }
  /* Word 274 "EC"/Length of EMC raw data */
  cc1='E';cc2='C';ss1=(short)(cc2<<8)+(short)cc1;
  ss2=480;uu=(ulong)(ss2<<16)+(ulong)ss1; write(gFd,&uu,4);
  /* Word 275->394 EMC RAW DATA */
  for (ss1=0;ss1<120;ss1++)
    {
       uu=gTrgd.RawEMC[ss1]; write(gFd,&uu,4);
    }
}
