 //Change log:  14-Jun-99 JLK (See README for details)

/***********************************************************************/
/* This program was developed under Linux/Pentium.  It may not
** be easily portable, although, of course, its output should be,
** since the DAQ banks were designed that way.
** For writing binary files in the DAQ data format.
** Page numbers in the comments below refer to the document
** "DAQ Data Format, revision 1.11". */
/******************************************  notes  ********************/
/*
  BBB things to do
  lint
  int (only for gFd)
  look out for negative ulongs!
  look for write(,,x) where x is not 4
  if( = )      /if([^\(\)]*[^=!<>]=[^=][^\(\)]*)/
  MB is frozen, upgrade to several.
  RECEIVERBOARDEVEN is frozen, upgrade to several

   This is comment 12e.  Instead of trying to spread the hits
   in a realistic fashion among the receiver boards (TPCRBP), I'm just
   going to put all the hits for the even sector into board number
   RECEIVERBOARDEVEN (counting from 1), and all the hits for the
   odd sector into board number RECEIVERBOARDODD.

*/
/******************************************  includes  *****************/
#ifndef MDATA_H
#include "mdata.h"
#endif
#ifndef FTP_H
#include "FTP.h"
#endif
/*********************************************  functions  *************/
void Err(int ln) {
  PP"Error :%d\n",ln); exit(1);
}

void InitGlobalsForTPCCPPR(void) {
  int ii; unsigned char *jelly;
  for(ii=0;ii<NUMDATAWORDSCPPR;ii++) gCPP[ii]=0xfc00fc00;
  jelly=(unsigned char*)&(gCPP[0]);
  jelly[0]=(0.2)*PULSEHEIGHT;
  jelly[1]=(0.8)*PULSEHEIGHT;
  jelly[2]=1; /* BBB ???  See page 14, data word 1 of TPCCPPR. */
  jelly[3]=1; /* BBB ???  See page 14, data word 1 of TPCCPPR. */
}
void PrepareGlobalsForTPCCPPR(ulong whichPad,ulong nCenterbins,ulong *cbin) {
  ulong basicOffset,ii; ushort *peanutButter;
  int beg,end;
  if(whichPad>=PADSPERSECTOR) ERR;
  basicOffset=32*whichPad+1; /* First word is diff for this bank. */
  for(ii=0;ii<nCenterbins;ii++) {
    if(ii+basicOffset>=NUMDATAWORDSCPPR) ERR;
    if(ii+basicOffset< 0               ) ERR;
    peanutButter=(ushort*)(&(gCPP[ii+basicOffset]));
    beg=cbin[ii]-PULSEWIDTH/2; if(beg<1) beg=1; if(beg>512) beg=512;
    end=cbin[ii]+PULSEWIDTH/2; if(end<1) end=1; if(end>512) end=512;
    peanutButter[0]=beg; peanutButter[1]=end;
  }
}
void WriteBodyTPCCPPR(char swapbytes,ulong junk) {
  write(gFd,gCPP,4*NUMDATAWORDSCPPR);
}

void Usage(void) {
  PP"  usage: %s [-s|-u] inputFile [triggerFile]\n",gExe);
  PP"example: %s -s events.input\n",gExe);
  PP"example: %s -u events.input\n",gExe);
  PP"-s means zero-suppressed data (including ADCD, SEQD, ADCX)\n");
  PP"-u means un-suppressed data (including ADCR, PADK)\n");
  PP"Here is a sample inputFile (# is a comment symbol):\n");
  PP"# detector 1 = TPC, detector 2 = FTP\n");
  PP"# event detector sector padrow pad centerbin\n");
  PP"    1      1       1      13    3    44\n");
  PP"    1      1       2      43   77    344\n");
  PP"    1      2       1       1    4    28\n");
  PP"    1      2       4       2   112   143\n");
  PP"    2      1       1      13    3    44\n");
  PP"    2      1       2      43   77    344\n");
  PP"    2      2       1       1    4    28\n");
  PP"    2      2       4       2   112   143\n");
  PP"\nThe triggerFile has the format below, the line order \n");
  PP"is not important, but the order within the line is\n");
  PP"The max value that can follow are L0(32) MZ(36) EC(120) the rest(64)\n");
  PP"Here is a sample triggerFile:\n");
  PP"LO: 1 2 3 4 5 6 \n");
  PP"L1: 1 2 3 4 5 6 \n");
  PP"L2: 1 2 3 4 5 6 \n");
  PP"MZ: 1 2 3 4 5 6 \n");
  PP"CT: 1 2 3 4 5 6 \n");
  PP"EC: 1 2 3 4 5 6 \n");
  exit(2);
}
void InitHitInfo(void) {
  int ii;
  gNevent=0;
  for(ii=0;ii<EVENT;ii++) { gEvent[ii].nhit=0; gFTPEvent[ii].nhit=0; }
}
void OrderednessError(ulong ln) {
  PP"-------------------------------------------------\n");
  PP"Aborted.  The input file not ordered.  Error is at line %d.\n",ln);
  PP"Each event    must be ordered by detector.\n");
  PP"Each detector must be ordered by sector.\n");
  PP"Each sector   must be ordered by padrow.\n");
  PP"Each padrow   must be ordered by pad.\n");
  PP"Each pad      must be ordered by centerbin.\n");
  PP"Hint: sort the file with the Unix command \"sort -n\".\n");
  PP"ABORTED.\n"); exit(2);
}

ulong Ooo(ulong xx) {
  ulong rv=1;
  if(xx>0) rv*=513; /* 512 time bins */
  if(xx>1) rv*=189; /* 188 pads */
  if(xx>2) rv*=46;  /* 45 padrows */
  if(xx>3) rv*=25;  /* 24 sectors */
  if(xx>4) rv*=2;   /* TPC = 1 */
  return rv;
}
void ReadInputFile(char *xx) {
  FILE *ff; char *cc,line[83]; int nhit,event,sector,padrow,pad,cbin;
  int det;
  ulong ordered,prevordered=0,ln=0;
  InitHitInfo();
  ff=fopen(xx,"r"); if(!ff) ERR;
  while(fgets(line,80,ff)) {
    ln++; if(*line=='#') continue;
    cc=strtok(line," \n"); if(!cc) ERR; event=atoi(cc);  
    cc=strtok(NULL," \n"); if(!cc) ERR; det=atoi(cc);
    cc=strtok(NULL," \n"); if(!cc) ERR; sector=atoi(cc);
    cc=strtok(NULL," \n"); if(!cc) ERR; padrow=atoi(cc);
    cc=strtok(NULL," \n"); if(!cc) ERR; pad=atoi(cc);
    cc=strtok(NULL," \n"); if(!cc) ERR; cbin=atoi(cc);
    cc=strtok(NULL," \n"); if( cc) ERR;
    if(gNevent<event) gNevent=event;
    if(event>EVENT) ERR; event--;
    if(det==1) {
      gDetectorPresence.TPC=1;
      ordered=cbin+Ooo(1)*pad+Ooo(2)*padrow+Ooo(3)*sector+
	      Ooo(4)*det+Ooo(5)*event;
      if(ordered<prevordered) OrderednessError(ln); prevordered=ordered;
      nhit=gEvent[event].nhit++; if(nhit>=HITSPEREVENT) ERR;
      gEvent[event].sector[nhit]    = sector;
      gEvent[event].row[nhit]    = padrow;
      gEvent[event].pad[nhit]       = pad;
      gEvent[event].cbin[nhit] = cbin;
    } 
    if (det==2) {
      gDetectorPresence.FTP=1;
      prevordered=0; /* zeroed for FTP */
      ordered=cbin+FTPOoo(1)*pad+FTPOoo(2)*padrow+FTPOoo(3)*sector+
	      FTPOoo(4)*det+FTPOoo(5)*event;
      if(ordered<prevordered) OrderednessError(ln); prevordered=ordered;
      nhit=gFTPEvent[event].nhit++; if(nhit>=HITSPEREVENT) ERR;
      gFTPEvent[event].sector[nhit]    = sector;
      gFTPEvent[event].row[nhit]    = padrow;
      gFTPEvent[event].pad[nhit]       = pad;
      gFTPEvent[event].cbin[nhit] = cbin;
    }     
  }
  fclose(ff);
}
void DumpGlobAndExit() {
  int ii,jj;
  if(gDetectorPresence.TPC) {
    for(ii=0;ii<gNevent;ii++) {
      PP"TPC Event %2d ------------------------------------------\n",ii+1);
      for(jj=0;jj<gEvent[jj].nhit;jj++) {
        PP"%5d %5d %5d %5d\n", gEvent[ii].sector[jj], gEvent[ii].row[jj],
            gEvent[ii].pad[jj], gEvent[ii].cbin[jj]);
      }
    }
  }
  if (gDetectorPresence.FTP) {
    for(ii=0;ii<gNevent;ii++) {
      PP"FTP Event %2d ------------------------------------------\n",ii+1);
      for(jj=0;jj<gFTPEvent[jj].nhit;jj++) {
        PP"%5d %5d %5d %5d\n", gFTPEvent[ii].sector[jj],
	  gFTPEvent[ii].row[jj], gFTPEvent[ii].pad[jj],
	  gFTPEvent[ii].cbin[jj]);
      }
    }
  }
  exit(2);
}

off_t Curr(int codingReminder) {
  if(codingReminder!=I) ERR;
  return lseek(gFd,0,SEEK_CUR);
}

void WriteString(char *string,ulong pad) {
  char oo; ulong ii;
  oo=0;
  if(strlen(string)>pad) ERR;
  for(ii=0;string[ii];ii++) write(gFd,string+ii,1);
  for(;ii<pad;ii++) write(gFd,&oo,1);
}

void WriteBodyTPCP(char swapbytes,ulong eventNum0) { /* p9 */
  ulong uu,ii;
  if(swapbytes) ERR;

  /* TPCP word 1-24 */
  uu=0; for(ii=0;ii<48;ii++) write(gFd,&uu,4); /* Offsets and lengths
  for non-empty hypersectors are written later. (comment 77y)*/

}
void WriteBodyTPCSECP(char swapbytes,ulong eventNum1) { /* p9 */
  ulong uu,ii;
  if(swapbytes) ERR;

  /* TPCSECP word 1-24 */
  uu=0; for(ii=0;ii<24;ii++) write(gFd,&uu,4); /* Offsets and lengths
  for non-empty hypersectors are written later. (comment 77z)*/

}
void WriteBodyDATAP(char swapbytes,ulong eventNum1) { /* p8 */
  ulong uu,ii;
  if(swapbytes) ERR;

  /* DATAP word 1 */
  gOffWordOneOfDATAPBody=Curr(I);
  uu=0; write(gFd,&uu,4); /* Come back and write this later, p8. */

  /* DATAP word 2 */
  uu=time(NULL); write(gFd,&uu,4);

  /* DATAP word 3 */
  uu=eventNum1; write(gFd,&uu,4);

  /* DATAP word 4 */
  uu=0; write(gFd,&uu,4);

  /* DATAP word 5 */
  uu=0; write(gFd,&uu,4);

  /* DATAP word 6 */
  if (gDetectorPresence.TPC && gDetectorPresence.FTP) uu=33;
  if (gDetectorPresence.TPC && !gDetectorPresence.FTP) uu=1;
  if (!gDetectorPresence.TPC && gDetectorPresence.FTP) uu=32;
  if (!gDetectorPresence.TPC && !gDetectorPresence.FTP) uu=0;
  write(gFd,&uu,4);

  /* DATAP word 7-128 */
  uu=0; for(ii=7;ii<=128;ii++) write(gFd,&uu,4); /* Write 7 and 8 later. */

}
void FillLongsForThisPad(ulong nCenterbins,ulong *cbin,ulong *data) {
  /* INTEGERS SO THAT NEG OUT OF BOUNDS VALUES CAN BE SKIPPED. */ int jj,kk;
  unsigned char *buf; ulong nn; int cnt=0;
  if(PULSEWIDTH%2!=1) ERR; /* must be odd so that centerbin is not ambiguous */
  nn=PULSEWIDTH/2;
  buf=(unsigned char*)data;
  for(jj=0;jj<512;jj++) buf[jj]=0;
  for(jj=0;jj<nCenterbins;jj++) { /* crude rectangular shape for pulses */
    for(kk=cbin[jj]-nn;kk<=cbin[jj]+nn;kk++) {
      if(kk<0||kk>=512) continue; /* time bucket number out of data range */
      buf[kk]+=PULSEHEIGHT;
      if(cnt++==PULSEWIDTH/2) buf[kk]+=5; /* Put a little peak at the center. */
    }
  }
}
void FillLongsForTPCPADK(ulong *data) { /* see nested struct page 12 */
  unsigned char *buf;
  ulong ii,rowIndex,padIndex;
  buf=(unsigned char*)data;
  for(ii=0;ii<192*4;ii++) buf[ii]=0xff;
  for(ii=0;ii<gNpad;ii++) {
    rowIndex=ii*2; padIndex=ii*2+1;
    if(padIndex>=192*4) ERR;
    buf[rowIndex]=gP[ii].row;
    buf[padIndex]=gP[ii].pad;
  }
}
void WriteBodyTPCPADK(char swapbytes,ulong eventNum0) { /* see page 12 */
  ulong ii,uu,data[192];
  if(swapbytes) ERR;

  /* TPCPADK word 1 */
  uu=512; /* see WriteBodyTPCADCR */ write(gFd,&uu,4);

  /* TPCPADK word 2-6 */
  uu=0; for(ii=2;ii<=6;ii++) write(gFd,&uu,4);

  FillLongsForTPCPADK(data);
  write(gFd,&data,4*192);
}
void WriteBodyTPCADCR(char swapbytes,ulong eventAndSector) { /* see page 12 */
  ulong pad[PADSPERSECTOR],row[PADSPERSECTOR],iC,ii;
  ulong data[128]; /* 4 ADC values per word (ie, per long) = 512 total, p 12 */
  ulong cbin[CENTERBIN],nCenterbins,iL,ihit,event,sector,nlist=0;
  char already;
  event=eventAndSector/1000;
  sector=eventAndSector-1000*event;
  if(swapbytes) ERR;
  if(TPCADCR_FORMAT!=0) ERR;  /* compression not implemented, see page 12 */

  /* mk lst of padrows/pads 4 this sector (to support 2 hits on same pad) */
  for(ihit=0;ihit<gEvent[event].nhit;ihit++) {
    if(gEvent[event].sector[ihit]!=sector) continue;
    already=0;
    for(iL=0;iL<nlist;iL++) {
      if(gEvent[event].row[ihit]==row[iL]&&gEvent[event].pad[ihit]==pad[iL]) {
        already=7; break;
      }
    }
    if(!already) {
      if(nlist>=PADSPERSECTOR) ERR;
      row[nlist]=gEvent[event].row[ihit];
      pad[nlist]=gEvent[event].pad[ihit];
      nlist++;
    }
  }

  /* now use the list (pad[] and row[]) */
  gNpad=0;
  InitGlobalsForTPCCPPR();
  for(iL=0;iL<nlist;iL++) { /* one loop per pad */
    /* Make list of centerbins from input file [see Usage()]. */
    nCenterbins=0;
    for(ihit=0;ihit<gEvent[event].nhit;ihit++) {
      if(gEvent[event].sector[ihit]!=sector) continue;
      if(gEvent[event].row[ihit]!=row[iL]) continue;
      if(gEvent[event].pad[ihit]!=pad[iL]) continue;
      if(nCenterbins>=CENTERBIN) ERR;
      cbin[nCenterbins++]=gEvent[event].cbin[ihit];
    }
    if(gNpad>=PAD) ERR; gP[gNpad].pad=pad[iL]; gP[gNpad].row=row[iL]; gNpad++;
    FillLongsForThisPad(nCenterbins,cbin,data);
    PrepareGlobalsForTPCCPPR(iL,nCenterbins,cbin);
    write(gFd,data,512);
  }

  /* Fill out the rest of the bank with zeros. */
  for(ii=0;ii<128;ii++) data[ii]=0;
  for(ii=0;ii<384-nlist;ii++) write(gFd,data,512); 
}
void WriteBodyTPCMZP(char swapbytes,ulong notUsed) {
  /* The length is variable (p 10), but we do no harm by writing
  ** all 24 words, and may avoid a bug later. */
  ulong ii,uu=0;
  if(swapbytes) ERR;
  for(ii=1;ii<=24;ii++) write(gFd,&uu,4);
}
void WriteBodyTPCRBP(char swapbytes,ulong notUsed) {
  ulong ii,uu=0;
  if(swapbytes) ERR;
  for(ii=1;ii<=22;ii++) write(gFd,&uu,4);
}
ulong SwapBytes(ulong xx) {
  ulong ii,rv; unsigned char *a,*b;
  a=(unsigned char*)&rv; b=(unsigned char*)&xx;
  for(ii=0;ii<4;ii++) a[ii]=b[3-ii];
  return rv;
}
void WriteBodyLHRD(char swapbytes,ulong arg) {
  ulong uu;
  if(!swapbytes) ERR;

  /* LHRD word 1 */
  uu=0x0f000000; write(gFd,&uu,4); 

  /* LHRD word 2 */
  uu=SwapBytes(1); write(gFd,&uu,4);

  /* LHRD word 3-4 */
       if(arg==1) WriteString("DATA",8);
  else if(arg==2) WriteString("ENDR",8);
  else if(arg==3) WriteString("BEGR",8);
  else ERR;

  /* LHRD word 5 */
  uu=0; write(gFd,&uu,4);
}

void WriteHeader(char swapbytes,char *type,ulong bankId) {
  ulong uu,ii;

  /* header word 1-2 */
  WriteString(type,8);

  /* header word 3 */
  uu=0; write(gFd,&uu,4); /* Filled in later at comment 44u. */

  /* header word 4 */
  write(gFd,&bankId,4); /* bytes already swapped in Bank(). */

  /* header word 5 (see word 7) */
  uu=1; write(gFd,&uu,4);

  /* header word 6 */
  uu=0x04030201; if(swapbytes) uu=SwapBytes(uu); write(gFd,&uu,4);

  /* header word 7 */
  if(!strcmp(type,"TPCADCR")) uu=TPCADCR_FORMAT;
  else if(!strcmp(type,"TPCADCD")) uu=TPCADCD_FORMAT;
  else if(!strcmp(type,"TPCP")) uu=1;
  else if(!strcmp(type,"TPCSECP")) uu=1;
/* added in FTP - only two format _requirements_ as of yet */
  else if(!strcmp(type,"FTPADCR")) uu=FTPADCR_FORMAT;
  else if(!strcmp(type,"FTPADCD")) uu=FTPADCD_FORMAT;
  else uu=0;
  write(gFd,&uu,4);

  /* header word 8-10 */
  uu=0; for(ii=8;ii<=10;ii++) write(gFd,&uu,4);

}
void WriteValueAtThisOffset(off_t off,ulong val) {
  off_t save;
  save=Curr(I);
  lseek(gFd,off,SEEK_SET);
  write(gFd,&val,4);
  lseek(gFd,save,SEEK_SET);
}
ulong ReadWordAsLong(off_t loc) {
  off_t save; ulong rv;
  save=Curr(I);
  lseek(gFd,loc,SEEK_SET);
  if(read(gFd,&rv,4)!=4) ERR;
  lseek(gFd,save,SEEK_SET);
  return rv;
}
/* comment 88n: These are non-zero only during the init, to reduce
** potential for confusion. */
void GetHitsThisSector(ulong event,ulong sector1,
      ulong *row,ulong *pad,ulong *cbin,ulong *nhits) {
  ulong ii,cnt=0;
  for(ii=0;ii<gEvent[event].nhit;ii++) {
    if(gEvent[event].sector[ii]!=sector1) continue;
    if(cnt>=HITSPERSECTOR) ERR;
    row[cnt]=gEvent[event].row[ii];
    pad[cnt]=gEvent[event].pad[ii];
    cbin[cnt]=gEvent[event].cbin[ii];
    cnt++;
  }
  *nhits=cnt;
}
ushort SwapTwoBytes(ushort input1) {
  ushort input2;
  char *cc1; void *uu1;
  char *cc2; void *uu2;
  uu1=&input1; cc1=uu1;
  uu2=&input2; cc2=uu2;
  cc2[0]=cc1[1];
  cc2[1]=cc1[0];
  return input2;
}
ushort Switch0(char lastHitThisPad,ulong cbin) {
  ushort rv,starttime;
  starttime=cbin-PULSEWIDTH/2;
  rv=PULSEWIDTH+32*(lastHitThisPad?1:0)+64*starttime; /* BBB comment 99z
                      ** Seq len not
                      ** correct if hit is near z=0 or z=TPCLEN/2 */
  /* May 15 1999 rv=SwapTwoBytes(rv); */
  return rv;
}
ushort Switch1(ulong row,ulong pad) {
  ushort rv;
  rv=0x8000+256*row+pad;
  return rv;
}
void WriteBodyTPCSEQD(char swapbytes,ulong arg) {
  ulong which73x,cnt=0,event0,sector1,ii; size_t nwords;
  ulong row[HITSPERSECTOR],pad[HITSPERSECTOR],cbin[HITSPERSECTOR];
  ulong body[MEM],ihit,nhit,packcnt=0;
  ushort *packed; void *silenceCompiler;
  char lastHitThisPad,newMarker,x0=0,x1=0,x2=0;
  silenceCompiler=body; packed=silenceCompiler; /* data packed 2/word */
  for(ii=0;ii<MEM;ii++) body[ii]=0;
  event0=arg/1000; sector1=arg-1000*event0;
  GetHitsThisSector(event0,sector1,row,pad,cbin,&nhit); gOtherBank=0;
  for(ihit=0;ihit<nhit;ihit++) {
    if(row[ihit]<1||row[ihit]>45) ERR;
    if(row[ihit]>= 1&&row[ihit]<=15&&!x0) {
      x0=7; gADCX[0].oSEQD=Curr(I)-gPh+ihit;
    }
    if(row[ihit]>=16&&row[ihit]<=30&&!x1) {
      x1=7; gADCX[1].oSEQD=Curr(I)-gPh+ihit;
    }
    if(row[ihit]>=31&&row[ihit]<=45&&!x2) {
      x2=7; gADCX[2].oSEQD=Curr(I)-gPh+ihit;
    }
    newMarker=0;
    if(ihit==0||row[ihit]!=row[ihit-1]) { /* new padrow */ newMarker=7; }
    else if(pad[ihit]>pad[ihit-1]+1) { /* skipped a pad */ newMarker=7; }
    lastHitThisPad=0;
    if(ihit<nhit-1) {
      if(pad[ihit]!=pad[ihit+1]) lastHitThisPad=7;
      if(row[ihit]!=row[ihit+1]) lastHitThisPad=7;
    } else lastHitThisPad=7;
    /* PP"event=%d, sector=%02d, row=%02d, pad=%02d, cb=%03d, %d %d\n",
    event0,sector1,row[ihit],pad[ihit],cbin[ihit],newMarker,lastHitThisPad); */
    if(newMarker) {
      if(packcnt>=2*MEM) ERR; packed[packcnt++]=Switch1(row[ihit],pad[ihit]);
    }
    if(packcnt>=2*MEM) ERR;
    packed[packcnt++]=Switch0(lastHitThisPad,cbin[ihit]); gOtherBank++;
  }
  nwords=(packcnt+1)/2; if(nwords>=MEM) ERR; write(gFd,body,4*nwords);
}
void WriteBodyTPCADCD(char swapbytes,ulong arg) { /* page 11 */
  ulong body[NMEM],row[HITSPERSECTOR],pad[HITSPERSECTOR],cbin[HITSPERSECTOR];
  ulong ii,nhit,ihit,event0,sector1,ndata=0; size_t nwords;
  unsigned char *data; void *b;
  char x0=0,x1=0,x2=0;
  event0=arg/1000; sector1=arg-1000*event0; b=body; data=b;
  GetHitsThisSector(event0,sector1,row,pad,cbin,&nhit);
  for(ii=0;ii<NMEM;ii++) body[ii]=0;
  for(ihit=0;ihit<nhit;ihit++) {
    if(row[ihit]<1||row[ihit]>45) ERR;
    if(row[ihit]>= 1&&row[ihit]<=15&&!x0) {
      x0=7; gADCX[0].oADCD=Curr(I)-gPh+ihit; gADCX[0].rowNumber=row[ihit];
    }
    if(row[ihit]>=16&&row[ihit]<=30&&!x1) {
      x1=7; gADCX[1].oADCD=Curr(I)-gPh+ihit; gADCX[1].rowNumber=row[ihit];
    }
    if(row[ihit]>=31&&row[ihit]<=45&&!x2) {
      x2=7; gADCX[2].oADCD=Curr(I)-gPh+ihit; gADCX[2].rowNumber=row[ihit];
    }
    for(ii=0;ii<PULSEWIDTH;ii++) {
      if(ndata>=NMEM*4) ERR;
      if(data[ndata]>=256-PULSEHEIGHT) ERR; data[ndata++]+=PULSEHEIGHT;
    }
  }
  if(ndata!=PULSEWIDTH*gOtherBank) ERR;
//  PP"bbb ndata=%d\n",ndata);
  nwords=(ndata+3)/4; if(nwords>=MEM) ERR; write(gFd,body,4*nwords);
}
void WriteBodyTPCADCX(char swapbytes,ulong arg) { /* page 13 */
  ulong val;
  int ii; /* int OK */
  for(ii=0;ii<3;ii++) {
    if(gADCX[ii].oADCD==0) continue;
    val=gADCX[ii].rowNumber; write(gFd,&val,4);
    val=gADCX[ii].oADCD-40;  write(gFd,&val,4); /* 40 bytes in header */
    val=gADCX[ii].oSEQD-40;  write(gFd,&val,4); /* 40 bytes in header */
  }
}
void Bank(char swapbytes,char *type, ulong bankId,void (*BodyFnct)(char,ulong),
      ulong argForBodyFnct,
      off_t oohopb,  /* Offset Of Head Of Parent Bank (bytes) */
      ulong wordNumOff1) {
  ulong wordNumLen1;
  off_t startHead; ulong lenBodyParBank,wordOffset,byteOffset,len;
  char lastLetter;
  wordNumLen1=wordNumOff1+1;
  lastLetter=type[strlen(type)-1];
  startHead=Curr(I); gPh=startHead;
  WriteHeader(swapbytes,type,swapbytes?SwapBytes(bankId):bankId);
  BodyFnct(swapbytes,argForBodyFnct);
  len=(Curr(I)-startHead)/4;
  if(!strcmp(type,"TPCCPPR")&&len!=NUMDATAWORDSCPPR+10) ERR; /* 12299 */
  if(!strcmp(type,"FTPCPPR")&&len!=NUMDATAWORDSFTPCPPR+10) ERR; /*10251*/
  WriteValueAtThisOffset(startHead+8,swapbytes?SwapBytes(len):len); /* 44u */
  if(oohopb>0) {
    lenBodyParBank=ReadWordAsLong(oohopb+2*4)-WHEADER;
    if(wordNumOff1>lenBodyParBank) ERR;
    if(wordNumLen1>lenBodyParBank) ERR;
    byteOffset=startHead-oohopb;
    if(byteOffset%4) ERR; wordOffset=byteOffset/4;
    if(wordNumOff1<1||wordNumLen1<1) ERR;
    /* The two writes below do not do a swappedness check. You can't use
    ** the swapbytes arg here, because it depends on parent bank.  As of
    ** May 3 1999, only the LRHD bank (not a parent) is swapped, so
    ** the lack of a swappedness check is an ignorable bug. */
    WriteValueAtThisOffset(oohopb+4*(WHEADER+wordNumOff1-1),wordOffset);
    WriteValueAtThisOffset(oohopb+4*(WHEADER+wordNumLen1-1),len);
  }
}
ulong NumHits(ulong ev0,ulong sector1) {
  ulong rv=0; long ii;
  for(ii=gEvent[ev0].nhit-1;ii>=0;ii--) {
    if(gEvent[ev0].sector[ii]==sector1) rv++;
  }
  return rv;
}
#define SUFFIX ".input"
void GenerateOutputFileName(char *oo,char *xx) {
  char *cc,*dd;
  strcpy(oo,xx);
  cc=strstr(oo,SUFFIX);
  if(!cc) {
    PP"Fatal error: the name of input the file must end in \".input\".\n");
    exit(2);
  }
  cc[0]=0;
  if(strstr(oo,SUFFIX)) ERR; /* Why do we have two SUFFIXs? */
  if(gSuppressed) strcat(cc,".mdata.sup");
  else            strcat(cc,".mdata.raw");
}
void ReadOption(char *xx) {
  if(xx[0]!='-') Usage();
  if(xx[1]=='s') gSuppressed=7;
  else if(xx[1]=='u') gSuppressed=0;
  else Usage();
}
void InitTheComment73xGlobals(void) {
  int ii;
  for(ii=0;ii<3;ii++) { gADCX[ii].oADCD=0; gADCX[ii].oSEQD=0; }
}
void CheckEndian(void) {
  ushort rr; void *cc; char *dd;
  cc=&rr; dd=cc;
  rr=0x1509;
  if(dd[0]!=9||dd[1]!=21) {
    PP"ERROR.  The machine you are trying to run on has the\n");
    PP"wrong endian-ness.  This program was developed under Linux/Pentium.\n");
    PP"If you want to port, it will take you a few hours.  Gameplan:\n");
    PP"compile and run the program on Linux/Pentium, make a baseline\n");
    PP"binary output file,\n");
    PP"and then hack the program until the output file on your \n");
    PP"new architecture\n");
    PP"matches the baseline.\n");
    exit(2);
  }
}
int main(int nnn,char *aaa[]) {
  char writeTrigBank,outname[188];
  ulong hsec; /* index receiver board, hyper sector */
  ulong FTPsec,chamber,FTPrb,extsec,low,high,rbhitcnt,chamberhitcnt,sechitcnt;
  ulong nbytes,rb,irb,hitcnt,ihit,ev,bid,ee,wordOff,wordOffset,sec1;
  off_t curr,bolr,oTPCSECP,oDATAP,oTPCP,oTPCRBP,oTPCMZP,oTRGD;
  off_t oFTPP,oFTPCHAP,oFTPRBP,oFTPAZIP,oFTPMZP;
  strcpy(gExe,aaa[0]);
  if((nnn<3)||(nnn>4)) Usage(); ReadOption(aaa[1]); writeTrigBank=0;
  if(sizeof(ulong)!=4) ERR; if(sizeof(ushort)!=2) ERR; CheckEndian();
  ReadInputFile(aaa[2]); GenerateOutputFileName(outname,aaa[2]);
  PP"Detectors present: TPC: %d, FTP: %d\n",gDetectorPresence.TPC,gDetectorPresence.FTP);
  InitTrigData();
  if (nnn==4) { if(ReadTrigData(aaa[3])) ERR; writeTrigBank=7; }
  /* DumpGlobAndExit(); */
  gFd=open(outname, O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if(gFd<0) ERR;
  Bank(7,"LRHD    ",123,WriteBodyLHRD,3,0,0);
  for(ev=0;ev<gNevent;ev++) {
    Bank(7,"LRHD",123,WriteBodyLHRD,1,0,0);
    oDATAP=Curr(I);
    Bank(0,"DATAP",0,WriteBodyDATAP,ev+1,0,0);
    if(writeTrigBank) {
      oTRGD=Curr(I);
      Bank(0,"TRGD",0,WriteBodyTRGD,0,oDATAP,23);
    }
    if (gDetectorPresence.TPC) {
      oTPCP=Curr(I);
      Bank(0,"TPCP",0,WriteBodyTPCP,ev,oDATAP,7);
      for(hsec=0;hsec<12;hsec++) {
        hitcnt=NumHits(ev,hsec*2+1)+NumHits(ev,hsec*2+2);
        if(hitcnt>0) { /* if hitcnt==0, leave the two zeros (comment 77y) */
          bid=hsec*2+1; wordOff=hsec*4+1; /* use 4 because of year 1 stuff */
          oTPCSECP=Curr(I);
          Bank(0,"TPCSECP",bid,WriteBodyTPCSECP,0,oTPCP,wordOff);
          for(irb=0;irb<2;irb++) { /* See comment 12e. */
            if(irb==0) { sec1=hsec*2+1; rb=RECEIVERBOARDODD;  }
            else       { sec1=hsec*2+2; rb=RECEIVERBOARDEVEN; }
            if((hitcnt=NumHits(ev,sec1))<1) continue;
            oTPCRBP=Curr(I);
            Bank(0,"TPCRBP",rb,WriteBodyTPCRBP,0,oTPCSECP,2*(rb-1)+1);
            oTPCMZP=Curr(I);
            Bank(0,"TPCMZP",MB,WriteBodyTPCMZP,0,oTPCRBP,2*(MB-1)+1);
            ee=1000*ev+sec1;
            if(gSuppressed) {
              InitTheComment73xGlobals();
              Bank(0,"TPCSEQD",MB,WriteBodyTPCSEQD,ee,oTPCMZP, 3); /* See */
              Bank(0,"TPCADCD",MB,WriteBodyTPCADCD,ee,oTPCMZP, 1); /* comment */
         /*   Bank(0,"TPCADCX", 0,WriteBodyTPCADCX,ee,oTPCMZP, 5); ** 73x. */
            } else {
              Bank(0,"TPCPADK",MB,WriteBodyTPCPADK, 0,oTPCMZP, 7);
              Bank(0,"TPCCPPR",MB,WriteBodyTPCCPPR, 0,oTPCMZP, 9);
              Bank(0,"TPCADCR",MB,WriteBodyTPCADCR,ee,oTPCMZP,11);
            }
          }
        }
      } 
    }  /* end write out TPC */
    if (gDetectorPresence.FTP) {
      oFTPP=Curr(I);
      Bank(0,"FTPP",0,WriteBodyFTPP,ev,oDATAP,17); /* 17: datapword for FTP offset */
      for(chamber=0;chamber<2;chamber++) {
        chamberhitcnt=0;  low=(chamber*30)+1;  high=(chamber*2)+30;
        for(extsec=low;extsec<=high;extsec++) {
	   chamberhitcnt+=NumFTPHits(ev,extsec);
        }
        if(chamberhitcnt>0) {
	  oFTPCHAP=Curr(I);
	  wordOffset=((chamber*2)+1);
	  Bank(0,"FTPCHAP",chamber,WriteBodyFTPCHAP,0,oFTPP,wordOffset);
	  for(FTPrb=0;FTPrb<10;FTPrb++) {
	    rbhitcnt=0;  low=(chamber*30)+(FTPrb*3)+1;  high=(chamber*30)+(FTPrb*3)+3;
	    for(extsec=low;extsec<=high;extsec++) {
              rbhitcnt+=NumFTPHits(ev,extsec);
            }
	    if(rbhitcnt>0) {      /* Write out non-empty receiver boards */
              oFTPRBP=Curr(I); 
              wordOffset=((FTPrb*2)+1);
              Bank(0,"FTPRBP",FTPrb,WriteBodyFTPRBP,0,oFTPCHAP,wordOffset);
              for(FTPsec=0;FTPsec<3;FTPsec++) {
	        sechitcnt=0; extsec=(chamber*30)+(FTPrb*3)+FTPsec+1;
                sechitcnt=NumFTPHits(ev,extsec);
                if(sechitcnt>0) {     /* Write out non-empty sector/mzboards */
                  oFTPAZIP=Curr(I);
                  wordOffset=((FTPsec*2)+1);
		  Bank(0,"FTPAZIP",FTPsec,WriteBodyFTPAZIP,0,oFTPRBP,wordOffset);
                  /* 1:1 sector to mz mapping */
	          oFTPMZP=Curr(I);
                  wordOffset=((FTPsec*2)+1);
		  Bank(0,"FTPMZP",FTPsec,WriteBodyFTPMZP,0,oFTPAZIP,wordOffset);
		  ee=1000*ev+extsec;
        	  if(gSuppressed) {
		    InitTheFTPComment73xGlobals();
		    Bank(0,"FTPSEQD",FTPsec,WriteBodyFTPSEQD,ee,oFTPMZP,3); /* See */
		    Bank(0,"FTPADCD",FTPsec,WriteBodyFTPADCD,ee,oFTPMZP,1); /* comment */
		    Bank(0,"FTPADCX",FTPsec,WriteBodyFTPADCX,ee,oFTPMZP,5); /* 73x */
		  } else {
		    Bank(0,"FTPADCR",FTPsec,WriteBodyFTPADCR,ee,oFTPMZP,11);
                    Bank(0,"FTPPADK",FTPsec,WriteBodyFTPPADK,ee,oFTPMZP, 7);
                    Bank(0,"FTPCPPR",FTPsec,WriteBodyFTPCPPR,ee,oFTPMZP, 9);   
		  }
		} /* endif sector not empty */
              }   /* end write-out non-empty sectors */
	    }     /* endif rcvboard not empty */
	  }       /* end write-out non-empty rcvboards */
	}         /* endif chamber not empty */
      }		  /* end write-out non-empty chambers */
    } 		  /* end write out FTP */

    nbytes=Curr(I)-oDATAP; if(nbytes%4) ERR;
    WriteValueAtThisOffset(oDATAP+4*10,nbytes/4); /* word 1 of DATAP body */

  }
  Bank(7,"LRHD",123,WriteBodyLHRD,2,0,0);
  close(gFd);
  PP"Normal end.  ");
  PP"Output is %s.\n",outname);
}
