#ifndef MDATA_H
#define MDATA_H

/* Standard includes */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>


/************  Multi-use abbreviations  *****************/
#define MEM 500
#define NMEM (MEM/2)
#define HITSPERSECTOR 100
#define PULSEWIDTH 27 /* units = time bins, full width */
#define PULSEHEIGHT 100
#define PAD 20
#define CENTERBIN 100 /* max hits per pad */
#define PADSPERSECTOR 50
#define I 4379
#define WHEADER 10  /* words in the header */
#define ERR Err(__LINE__)
#define PP printf(

/************  TPC (mdata.c) abbreviations  *************/
#define NUMDATAWORDSCPPR 12289
#define TPCADCR_FORMAT 0
#define TPCADCD_FORMAT 0 /* uncompressed */
#define MB 2  /* For simplicity, put all data into this one mezz board. */
#define RECEIVERBOARDEVEN 9 /* Reflects changes in v1.13 of doc. */
#define RECEIVERBOARDODD 3  /* Reflects changes in v1.13 of doc. */

/************  enumerations  ******************/
/**************  Multi-use static memory  *****/
#define HITSPEREVENT 30
#define EVENT 10

/**************  Multi-use typedefs  **********/
typedef struct {
  int TPC,FTP;
} DET_PRES;

/**************  TPC typedefs  ****************/
typedef struct {
  int pad,row;
} PAD_T;
typedef struct {
  int sector[HITSPEREVENT];
  int row[HITSPEREVENT];
  int pad[HITSPEREVENT];
  int cbin[HITSPEREVENT];
  int nhit;
} EVENT_T;
typedef struct {
  ulong oADCD,oSEQD;
  int rowNumber;
} TPCADCX_T;
  
/************  navigation  ********************/
/************  Multi-use globals  *************/
char gExe[200];
int gNevent;
off_t gOffWordOneOfLRHDBody;
off_t gOffWordOneOfDATAPBody;
off_t gPh;
int gFd;
char gSuppressed;
DET_PRES gDetectorPresence;

/************  TPC (mdata.c) globals  *********/ 
ulong gCPP[NUMDATAWORDSCPPR];
int gOtherBank; /* int is OK */
EVENT_T gEvent[EVENT];
PAD_T gP[PAD];
                    /* This distribution scheme may be flawed. */
TPCADCX_T gADCX[3]; /* rows 1-15 to [0], 16-30 to [1], and 31-45 to [2]. */

ulong gNumTPCADCX;
int gNpad;

/************  Multi-use functions  ***********/
void Err(int ln);
void Usage(void);
void OrderednessError(ulong ln);
void ReadInputFile(char *xx);
void DumpGlobAndExit();
off_t Curr(int codingReminder);
void WriteString(char *string,ulong pad);
void WriteBodyDATAP(char swapbytes,ulong eventNum1);
ulong SwapBytes(ulong xx);
void WriteBodyLHRD(char swapbytes,ulong arg);
void WriteHeader(char swapbytes,char *type,ulong bankId);
void WriteValueAtThisOffset(off_t off,ulong val);
ulong ReadWordAsLong(off_t loc);
ushort SwapTwoBytes(ushort input1);
ushort Switch0(char lastHitThisPad,ulong cbin);
ushort Switch1(ulong row,ulong pad);
void Bank(char swapbytes,char *type, ulong bankId,void
(*BodyFnct)(char,ulong),
      ulong argForBodyFnct, 
      off_t oohopb,  /* Offset Of Head Of Parent Bank (bytes) */
      ulong wordNumOff1);
void GenerateOutputFileName(char *oo,char *xx);
void ReadOption(char *xx);
void CheckEndian(void);

/************  TPC (mdata.c) functions  *******/
void InitHitInfo(void);
void InitGlobalsForTPCCPPR(void);
void PrepareGlobalsForTPCCPPR(ulong whichPad,ulong nCenterbins,ulong *cbin);
void WriteBodyTPCCPPR(char swapbytes,ulong junk);
ulong Ooo(ulong xx);
void WriteBodyTPCP(char swapbytes,ulong eventNum0);
void WriteBodyTPCSECP(char swapbytes,ulong eventNum1);
void FillLongsForThisPad(ulong nCenterbins,ulong *cbin,ulong *data);
void FillLongsForTPCPADK(ulong *data);
void WriteBodyTPCPADK(char swapbytes,ulong eventNum0);
void WriteBodyTPCADCR(char swapbytes,ulong eventAndSector);
void WriteBodyTPCMZP(char swapbytes,ulong notUsed);
void WriteBodyTPCRBP(char swapbytes,ulong notUsed);
void GetHitsThisSector(ulong event,ulong sector1,
      ulong *row,ulong *pad,ulong *cbin,ulong *nhits);
void WriteBodyTPCSEQD(char swapbytes,ulong arg);
void WriteBodyTPCADCD(char swapbytes,ulong arg);
void WriteBodyTPCADCX(char swapbytes,ulong arg);
ulong NumHits(ulong ev0,ulong sector1);
void InitTheComment73xGlobals(void);

/************  TRG (dan.c) functions  *******/
void InitTrigData(void);
int ReadTrigData(char *fname);
void WriteBodyTRGD(char swapbytes,ulong arg);


#endif
